<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform" xmlns:vis="http://schemas.microsoft.com/visio/2003/core" xmlns="http://www.w3.org/1999/XSL/Transform" xmlns:xs="http://www.w3.org/2001/XMLSchema" xmlns:fdo="http://www.autodesk.com/isd/fdo" xmlns:gml="http://www.opengis.net/gml">
  <!--<xsl:output method="xml" version="1.0" encoding="UTF-8" indent="yes"/>-->

  <!-- Store stereotype IDs in global variables: -->
  <xsl:variable name="featStereo" select="//Foundation.Core.Namespace.ownedElement/Foundation.Extension_Mechanisms.Stereotype[Foundation.Core.ModelElement.name = 'Feature']/@xmi.id"/>
  <xsl:variable name="identStereo" select="//Foundation.Core.Namespace.ownedElement/Foundation.Extension_Mechanisms.Stereotype[Foundation.Core.ModelElement.name = 'Ident']/@xmi.id"/>

  <!-- Network stereotype IDs ; not yet supported:
  <xsl:variable name="networkStereo" select="//Foundation.Core.Namespace.ownedElement/Foundation.Extension_Mechanisms.Stereotype[Foundation.Core.ModelElement.name = 'Network']/@xmi.id"/>
  <xsl:variable name="networkLayerStereo" select="//Foundation.Core.Namespace.ownedElement/Foundation.Extension_Mechanisms.Stereotype[Foundation.Core.ModelElement.name = 'NetworkLayer']/@xmi.id"/>
  <xsl:variable name="networkLinkFeatureStereo" select="//Foundation.Core.Namespace.ownedElement/Foundation.Extension_Mechanisms.Stereotype[Foundation.Core.ModelElement.name = 'NetworkLinkFeature']/@xmi.id"/>
  <xsl:variable name="networkNodeFeatureStereo" select="//Foundation.Core.Namespace.ownedElement/Foundation.Extension_Mechanisms.Stereotype[Foundation.Core.ModelElement.name = 'NetworkNodeFeature']/@xmi.id"/>
  <xsl:variable name="networkAssociationStereo" select="//Foundation.Core.Namespace.ownedElement/Foundation.Extension_Mechanisms.Stereotype[Foundation.Core.ModelElement.name = 'NetworkAssociation']/@xmi.id"/>
  <xsl:variable name="referencedFeatureAssociationStereo" select="//Foundation.Core.Namespace.ownedElement/Foundation.Extension_Mechanisms.Stereotype[Foundation.Core.ModelElement.name = 'ReferencedFeatureAssociation']/@xmi.id"/>
  <xsl:variable name="parentNetworkFeatureAssociationStereo" select="//Foundation.Core.Namespace.ownedElement/Foundation.Extension_Mechanisms.Stereotype[Foundation.Core.ModelElement.name = 'ParentNetworkFeatureAssociation']/@xmi.id"/>
  <xsl:variable name="startNodeAssociationStereo" select="//Foundation.Core.Namespace.ownedElement/Foundation.Extension_Mechanisms.Stereotype[Foundation.Core.ModelElement.name = 'StartNodeAssociation']/@xmi.id"/>
  <xsl:variable name="endNodeAssociationStereo" select="//Foundation.Core.Namespace.ownedElement/Foundation.Extension_Mechanisms.Stereotype[Foundation.Core.ModelElement.name = 'EndNodeAssociation']/@xmi.id"/>
  <xsl:variable name="layerAssociationStereo" select="//Foundation.Core.Namespace.ownedElement/Foundation.Extension_Mechanisms.Stereotype[Foundation.Core.ModelElement.name = 'LayerAssociation']/@xmi.id"/>
  -->


  <xsl:template match="XMI">
    <xsl:apply-templates/>
  </xsl:template>

  
  <xsl:template match="XMI.header">
  </xsl:template>

  
  <xsl:template match="XMI.content">
    <xsl:apply-templates select="*"/>
  </xsl:template>

  
  <!-- Do I need to allow for Foundation.Core.Namespace.ownedElement intermediate containment element? -->
  <xsl:template match="Model_Management.Subsystem">
    <xsl:call-template name="WriteSchema"/>
  </xsl:template>

  
  <!-- Do I need to allow for Foundation.Core.Namespace.ownedElement intermediate containment element? -->
  <xsl:template match="Model_Management.Model">
    <xsl:call-template name="WriteSchema"/>
  </xsl:template>

  
  <xsl:template match="Foundation.Core.Namespace.ownedElement">
    <xsl:apply-templates select="*"/>
  </xsl:template>


  <xsl:template match="Model_Management.Package">
    <xsl:call-template name="WriteSchema"/>
  </xsl:template>

  
  <!-- Everything else gets thrown away -->
  <xsl:template match="*" priority="-1">
    <!--    <xsl:message> throwing away  <xsl:value-of select="Foundation.Core.ModelElement.name"/> </xsl:message>-->
  </xsl:template>



  
  <xsl:template name="WriteSchema">
    <!-- Don't create a Schema element if this is a *nested* Package/Subsystem/Model -->
    <xsl:choose>

      <xsl:when test="ancestor::Model_Management.Package or ancestor::Model_Management.Subsystem or ancestor::Model_Management.Model">
        <xsl:call-template name="WriteClasses"/>
        <xsl:apply-templates/>
      </xsl:when>

      <xsl:otherwise>
        <xsl:element name="Schema">
          <xsl:variable name="fdoSchemaName">
            <xsl:call-template name="FormatName">
              <xsl:with-param name="inName" select="Foundation.Core.ModelElement.name"/>
            </xsl:call-template>
          </xsl:variable>
          
          <xsl:message>XmiToFdo: Mapping XMI Package/Subsystem/Model '<xsl:value-of select="Foundation.Core.ModelElement.name"/>' to FDO Schema '<xsl:value-of select="$fdoSchemaName"/>'</xsl:message>
 
          <xsl:call-template name="WriteTaggedAttributes"/>
          <xsl:attribute name="name">
            <xsl:value-of select="$fdoSchemaName"/>
          </xsl:attribute>

          <xsl:call-template name="WriteTaggedDocumentation"/>
          <xsl:call-template name="WriteClasses"/>
          <xsl:apply-templates/>
        </xsl:element>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>


  <xsl:template name="WriteClasses">
    <xsl:for-each select="Foundation.Core.Namespace.ownedElement/Foundation.Core.Class">
      <xsl:call-template name="WriteClass"/>
    </xsl:for-each>
  </xsl:template>


  <xsl:template name="WriteClass">
    <xsl:variable name="classType">
      <xsl:call-template name="GetClassType"/>
    </xsl:variable>
    <xsl:element name="{$classType}">
      <xsl:variable name="fdoClassName">
        <xsl:call-template name="FormatName">
          <xsl:with-param name="inName" select="Foundation.Core.ModelElement.name"/>
        </xsl:call-template>
      </xsl:variable>

      <xsl:message>XmiToFdo: mapping XMI Class '<xsl:value-of select="Foundation.Core.ModelElement.name"/>' to FDO <xsl:value-of select="$classType"/> '<xsl:value-of select="$fdoClassName"/>'</xsl:message>

      <xsl:attribute name="name">
        <xsl:value-of select="$fdoClassName"/>
      </xsl:attribute>
      <xsl:call-template name="WriteBaseClassAttributes"/>
      <xsl:call-template name="WriteTaggedAttributes"/>
      <xsl:choose>
        <xsl:when test="$classType='FeatureClass'">
          <xsl:variable name="geometryProperty">
            <xsl:call-template name="GetGeometryProperty"/>
          </xsl:variable>
          <xsl:choose>
            <xsl:when test="not(string-length($geometryProperty) = 0)">
              <xsl:attribute name="geometryProperty">
                <xsl:value-of select="$geometryProperty"/>
              </xsl:attribute>
              <xsl:attribute name="hasGeometry">true</xsl:attribute>
            </xsl:when>
            <xsl:otherwise>
              <xsl:attribute name="hasGeometry">false</xsl:attribute>
            </xsl:otherwise>
          </xsl:choose>
        </xsl:when>
        
        <!-- Network class-types; not yet suported:
        <xsl:when test="$classType='NetworkClass'">
        </xsl:when>
        <xsl:when test="$classType='NetworkLayerClass'">
        </xsl:when>
        <xsl:when test="$classType='NetworkLinkFeatureClass'">
          <xsl:call-template name="WriteGenericNetworkFeatureClassAttributes"/>
        </xsl:when>
        <xsl:when test="$classType='NetworkNodeFeatureClass'">
          <xsl:call-template name="WriteGenericNetworkFeatureClassAttributes"/>
        </xsl:when>
        -->
      </xsl:choose>
      <xsl:for-each select="Foundation.Core.GeneralizableElement.isAbstract">
        <xsl:attribute name="abstract">
          <xsl:value-of select="@xmi.value"/>
        </xsl:attribute>
      </xsl:for-each>

      <xsl:call-template name="WriteTaggedDocumentation"/>
      <xsl:call-template name="WriteIdentityProperties"/>
      <xsl:call-template name="WriteProperties"/>
      <xsl:choose>
        <!-- Network class types; not yet supported:
        <xsl:when test="$classType='NetworkClass'">
        </xsl:when>
        <xsl:when test="$classType='NetworkLayerClass'">
        </xsl:when>
        <xsl:when test="$classType='NetworkLinkFeatureClass'">
          <xsl:call-template name="WriteGenericNetworkFeatureClassElements"/>
        </xsl:when>
        <xsl:when test="$classType='NetworkNodeFeatureClass'">
          <xsl:call-template name="WriteGenericNetworkFeatureClassElements"/>
        </xsl:when>
        -->
      </xsl:choose>
      
    </xsl:element>
  </xsl:template>


  <!-- Network related templates; not yet supported:
  <xsl:template name="WriteGenericNetworkFeatureClassAttributes">
  </xsl:template>
  <xsl:template name="WriteGenericNetworkFeatureClassElements">
    <xsl:variable name="NetworkAssociationEnd">
      <xsl:call-template name="GetStereotypedAssociationEnd">
        <xsl:with-param name="stereotypeid" select="$networkAssociationStereo"/>
      </xsl:call-template>
    </xsl:variable>
    <xsl:if test="$NetworkAssociationEnd">
      <xsl:element name="NetworkProperty">
        <xsl:for-each select="$NetworkAssociationEnd">
          <xsl:call-template name="WriteAssociationProperty"/>
        </xsl:for-each>
      </xsl:element>
    </xsl:if>
    <xsl:variable name="ReferencedFeatureAssociationEnd">
      <xsl:call-template name="GetStereotypedAssociationEnd">
        <xsl:with-param name="stereotypeid" select="$referencedFeatureAssociationStereo"/>
      </xsl:call-template>
    </xsl:variable>
    <xsl:if test="$ReferencedFeatureAssociationEnd">
      <xsl:element name="ReferenceFeatureProperty">
        <xsl:for-each select="$ReferencedFeatureAssociationEnd">
          <xsl:call-template name="WriteAssociationProperty"/>
        </xsl:for-each>
      </xsl:element>
    </xsl:if>
    <xsl:variable name="ParentNetworkFeatureAssociationEnd">
      <xsl:call-template name="GetStereotypedAssociationEnd">
        <xsl:with-param name="stereotypeid" select="$parentNetworkFeatureAssociationStereo"/>
      </xsl:call-template>
    </xsl:variable>
    <xsl:if test="$ParentNetworkFeatureAssociationEnd">
      <xsl:element name="ParentNetworkProperty">
        <xsl:for-each select="$ParentNetworkFeatureAssociationEnd">
          <xsl:call-template name="WriteAssociationProperty"/>
        </xsl:for-each>
      </xsl:element>
    </xsl:if>
  </xsl:template>
  -->


  <xsl:template name="WriteTaggedAttributes">
    <!-- Note: Visio and ArgoUml represent multiple TaggedValues on an element differently:
        Visio: creates 1 Foundation.Core.ModelElement.taggedValue element per tag
        ArgoUml: creates 1 Foundation.Core.ModelElement.taggedValue for all tags, 
           then 1 Foundation.Extension_Mechanisms.TaggedValue subelement per tag.  -->
    <xsl:for-each select="Foundation.Core.ModelElement.taggedValue">
      <xsl:for-each select="Foundation.Extension_Mechanisms.TaggedValue">
        <xsl:variable name="attName" select="Foundation.Extension_Mechanisms.TaggedValue.tag"/>
        <xsl:variable name="attValue" select="Foundation.Extension_Mechanisms.TaggedValue.value"/>
        <xsl:if test="$attName != 'documentation' and string-length($attValue) != 0">
          <xsl:attribute name="{$attName}">
            <xsl:value-of select="$attValue"/>
          </xsl:attribute>
        </xsl:if>
      </xsl:for-each>
    </xsl:for-each>
  </xsl:template>


  <xsl:template name="WriteTaggedDocumentation">
    <xsl:for-each select="Foundation.Core.ModelElement.taggedValue">
      <xsl:for-each select="Foundation.Extension_Mechanisms.TaggedValue">
        <xsl:variable name="attName" select="Foundation.Extension_Mechanisms.TaggedValue.tag"/>
        <xsl:variable name="attValue" select="Foundation.Extension_Mechanisms.TaggedValue.value"/>
        <xsl:if test="$attName = 'documentation' and string-length($attValue) != 0">
          <xsl:element name="Description">
            <xsl:value-of select="$attValue"/>
          </xsl:element>
        </xsl:if>
      </xsl:for-each>
    </xsl:for-each>
  </xsl:template>

  
  <xsl:template name="WriteReadOnlyAttribute">
    <xsl:param name="attrName"/>
    <xsl:variable name="changeable" select="Foundation.Core.StructuralFeature.changeable/@xmi.value"/>
    <xsl:if test="$changeable">
      <xsl:attribute name="{$attrName}">
        <xsl:choose>
          <xsl:when test="$changeable='frozen'">true</xsl:when>
          <xsl:otherwise>false</xsl:otherwise>
        </xsl:choose>
      </xsl:attribute>
    </xsl:if>
  </xsl:template>

  
  <xsl:template name="GetMinMultiplicity">
    <xsl:call-template name="FormatMultiplicity">
      <xsl:with-param name="inMultiplicity">
        <xsl:choose>

          <!-- ArgoUml allows a Multiplicity to reference another one anywhere within the document: -->
          <xsl:when test="Foundation.Data_Types.Multiplicity/@xmi.idref">
            <xsl:variable name="MultiplicityId" select="Foundation.Data_Types.Multiplicity/@xmi.idref"/>
            <xsl:for-each select="//Foundation.Data_Types.Multiplicity[@xmi.id=$MultiplicityId]/..">
              <xsl:call-template name="GetMinMultiplicity"/>
            </xsl:for-each>
          </xsl:when>

          <xsl:when test="Foundation.Data_Types.Multiplicity/Foundation.Data_Types.Multiplicity.range/Foundation.Data_Types.MultiplicityRange/Foundation.Data_Types.MultiplicityRange.lower/text()">
            <xsl:value-of select="Foundation.Data_Types.Multiplicity/Foundation.Data_Types.Multiplicity.range/Foundation.Data_Types.MultiplicityRange/Foundation.Data_Types.MultiplicityRange.lower/text()"/>
          </xsl:when>

          <!-- Visio's approach to multiplicity ranges: -->
          <xsl:when test="text()='0..*'">0</xsl:when>
          <xsl:when test="text()='0..1'">0</xsl:when>
          <xsl:when test="text()='1..*'">1</xsl:when>
          <xsl:when test="text()='1..1'">1</xsl:when>

          <xsl:when test="text()">
            <xsl:value-of select="text()"/>
          </xsl:when>
          
        </xsl:choose>
      </xsl:with-param>
    </xsl:call-template>
  </xsl:template>

  
  <xsl:template name="GetMaxMultiplicity">
    <xsl:call-template name="FormatMultiplicity">
      <xsl:with-param name="inMultiplicity">
        <xsl:choose>

          <!-- ArgoUml allows a Multiplicity to reference another one anywhere within the document: -->
          <xsl:when test="Foundation.Data_Types.Multiplicity/@xmi.idref">
            <xsl:variable name="MultiplicityId" select="Foundation.Data_Types.Multiplicity/@xmi.idref"/>
            <xsl:for-each select="//Foundation.Data_Types.Multiplicity[@xmi.id=$MultiplicityId]/..">
              <xsl:call-template name="GetMaxMultiplicity"/>
            </xsl:for-each>
          </xsl:when>

          <xsl:when test="Foundation.Data_Types.Multiplicity/Foundation.Data_Types.Multiplicity.range/Foundation.Data_Types.MultiplicityRange/Foundation.Data_Types.MultiplicityRange.upper/text()">
            <xsl:value-of select="Foundation.Data_Types.Multiplicity/Foundation.Data_Types.Multiplicity.range/Foundation.Data_Types.MultiplicityRange/Foundation.Data_Types.MultiplicityRange.upper/text()"/>
          </xsl:when>

          <!-- Visio's approach to multiplicity ranges: -->
          <xsl:when test="text()='0..*'">m</xsl:when>
          <xsl:when test="text()='0..1'">1</xsl:when>
          <xsl:when test="text()='1..*'">m</xsl:when>
          <xsl:when test="text()='1..1'">1</xsl:when>

          <xsl:when test="text()">
            <xsl:value-of select="text()"/>
          </xsl:when>
          
        </xsl:choose>
      </xsl:with-param>
    </xsl:call-template>
  </xsl:template>

  
  <xsl:template name="WriteNullableAttribute">
    <xsl:variable name="minMultiplicity">
      <xsl:for-each select="Foundation.Core.StructuralFeature.multiplicity">
        <xsl:call-template name="GetMinMultiplicity"/>
      </xsl:for-each>
    </xsl:variable>
    <xsl:if test="$minMultiplicity = '0'">
      <xsl:attribute name="nullable">true</xsl:attribute>
    </xsl:if>
  </xsl:template>

  
  <xsl:template name="WriteIdentityProperties">
    <xsl:choose>
      <!-- Visio's way of storing stereotypes: -->
      <xsl:when test="Foundation.Core.Classifier.feature/Foundation.Core.Attribute[Foundation.Core.ModelElement.stereotype/Foundation.Extension_Mechanisms.Stereotype/@xmi.idref=$identStereo]">
        <xsl:element name="IdentityProperties">
          <xsl:for-each select="Foundation.Core.Classifier.feature/Foundation.Core.Attribute[Foundation.Core.ModelElement.stereotype/Foundation.Extension_Mechanisms.Stereotype/@xmi.idref=$identStereo]">
            <xsl:call-template name="WriteIdentityProperty"/>
          </xsl:for-each>
        </xsl:element>
      </xsl:when>
      <!-- ArgoUML's way of storing stereotypes: -->
      <xsl:when test="Foundation.Core.Classifier.feature/Foundation.Core.Attribute[@xmi.id = //Foundation.Core.Namespace.ownedElement/Foundation.Extension_Mechanisms.Stereotype[@xmi.id = $identStereo]/Foundation.Extension_Mechanisms.Stereotype.extendedElement/Foundation.Core.ModelElement/@xmi.idref]">
        <xsl:element name="IdentityProperties">
          <xsl:for-each select="Foundation.Core.Classifier.feature/Foundation.Core.Attribute[@xmi.id = //Foundation.Core.Namespace.ownedElement/Foundation.Extension_Mechanisms.Stereotype[@xmi.id = $identStereo]/Foundation.Extension_Mechanisms.Stereotype.extendedElement/Foundation.Core.ModelElement/@xmi.idref]">
            <xsl:call-template name="WriteIdentityProperty"/>
          </xsl:for-each>
        </xsl:element>
      </xsl:when>
      <xsl:otherwise>
        <xsl:call-template name="WriteBaseIdentityProperties"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>


  <xsl:template name="WriteBaseIdentityProperties">
    <xsl:variable name="baseClassId"><xsl:call-template name="GetBaseClassId"/></xsl:variable>
    <xsl:for-each select="//Foundation.Core.Namespace.ownedElement/Foundation.Core.Class[@xmi.id = $baseClassId]">
      <xsl:call-template name="WriteIdentityProperties"/>
    </xsl:for-each>
  </xsl:template>


  <xsl:template name="WriteIdentityProperty">
    <xsl:element name="IdentityProperty">
      <xsl:value-of select="Foundation.Core.ModelElement.name"/>
    </xsl:element>
  </xsl:template>

  
  <xsl:template name="WriteProperties">
    <xsl:element name="Properties">
      <!-- Write non-association properties: -->
      <xsl:for-each select="Foundation.Core.Classifier.feature/Foundation.Core.Attribute">
        <xsl:call-template name="WriteProperty"/>
      </xsl:for-each>
      
      <!-- Write association properties: -->
      <xsl:variable name="foreignId" select="@xmi.id"/>
      <xsl:for-each select="//Foundation.Core.Namespace.ownedElement/Foundation.Core.Association/Foundation.Core.Association.connection/Foundation.Core.AssociationEnd[Foundation.Core.AssociationEnd.type/child::*[self::Foundation.Core.Class or self::Foundation.Core.Classifier]/@xmi.idref=$foreignId]">
        <xsl:call-template name="WriteAssociationProperty"/>
      </xsl:for-each>
    </xsl:element>
  </xsl:template>

  
  <xsl:template name="WriteProperty">

    <xsl:variable name="attrTypeId">
      <xsl:choose>
        <!-- Visio's way for data attributes: -->
        <xsl:when test="Foundation.Core.StructuralFeature.type/Foundation.Core.DataType/@xmi.idref">
          <xsl:value-of select="Foundation.Core.StructuralFeature.type/Foundation.Core.DataType/@xmi.idref"/>
        </xsl:when>
        <!-- Visio's way for class attributes: -->
        <xsl:when test="Foundation.Core.StructuralFeature.type/Foundation.Core.Class/@xmi.idref">
          <xsl:value-of select="Foundation.Core.StructuralFeature.type/Foundation.Core.Class/@xmi.idref"/>
        </xsl:when>
        <!-- ArgoUML's way for data and class attributes: -->
        <xsl:when test="Foundation.Core.StructuralFeature.type/Foundation.Core.Classifier/@xmi.idref">
          <xsl:value-of select="Foundation.Core.StructuralFeature.type/Foundation.Core.Classifier/@xmi.idref"/>
        </xsl:when>
      </xsl:choose>
    </xsl:variable>

    <xsl:variable name="fdoPropertyName">
      <xsl:call-template name="FormatName">
        <xsl:with-param name="inName" select="Foundation.Core.ModelElement.name"/>
      </xsl:call-template>
    </xsl:variable>

    <xsl:choose>
      <xsl:when test="//Foundation.Core.DataType[@xmi.id = $attrTypeId]">

        <xsl:variable name="dataType">
          <xsl:for-each select="//Foundation.Core.DataType[@xmi.id = $attrTypeId]">
            <xsl:call-template name="getDataType">
              <xsl:with-param name="xmiType" select="Foundation.Core.ModelElement.name"/>
            </xsl:call-template>
          </xsl:for-each>
        </xsl:variable>

        <xsl:choose>

          <xsl:when test="$dataType='Geometric'">

            <xsl:message>XmiToFdo: mapping XMI Attribute '<xsl:value-of select="Foundation.Core.ModelElement.name"/>' to FDO Geometric Property '<xsl:value-of select="$fdoPropertyName"/>'</xsl:message>

            <xsl:element name="GeometricProperty">
              <xsl:attribute name="name">
                <xsl:value-of select="$fdoPropertyName"/>
              </xsl:attribute>
              <xsl:variable name="ElevationSpecified">
                <xsl:call-template name="TaggedAttributeExists">
                  <xsl:with-param name="tagName" select="'hasElevation'"/>
                </xsl:call-template>
              </xsl:variable>
              <xsl:if test="$ElevationSpecified='false'">
                <xsl:attribute name="hasElevation">true</xsl:attribute>
              </xsl:if>
              <xsl:call-template name="WriteTaggedAttributes"/>
              <xsl:call-template name="WriteReadOnlyAttribute">
                <xsl:with-param name="attrName" select="'readOnly'"/>
              </xsl:call-template>

              <xsl:call-template name="WriteTaggedDocumentation"/>
            </xsl:element>
          </xsl:when>

          <xsl:when test="$dataType='Raster'">

            <xsl:message>XmiToFdo: mapping XMI Attribute '<xsl:value-of select="Foundation.Core.ModelElement.name"/>' to FDO Raster Property '<xsl:value-of select="$fdoPropertyName"/>'</xsl:message>
            
            <xsl:element name="RasterProperty">
              <xsl:attribute name="name">
                <xsl:value-of select="$fdoPropertyName"/>
              </xsl:attribute>
              <xsl:call-template name="WriteReadOnlyAttribute">
                <xsl:with-param name="attrName" select="'readOnly'"/>
              </xsl:call-template>
              <xsl:call-template name="WriteNullableAttribute"/>
              <xsl:call-template name="WriteTaggedAttributes"/>

              <xsl:call-template name="WriteTaggedDocumentation"/>
            </xsl:element>
          </xsl:when>

          <xsl:otherwise>

            <xsl:message>XmiToFdo: mapping XMI Attribute '<xsl:value-of select="Foundation.Core.ModelElement.name"/>' to FDO Data Property '<xsl:value-of select="$fdoPropertyName"/>'</xsl:message>

            <xsl:element name="DataProperty">
              <xsl:attribute name="name">
                <xsl:value-of select="$fdoPropertyName"/>
              </xsl:attribute>
              <xsl:call-template name="WriteNullableAttribute"/>
              <xsl:call-template name="WriteTaggedAttributes"/>
              <xsl:attribute name="dataType">
                <xsl:value-of select="$dataType"/>
              </xsl:attribute>
              <xsl:variable name="LengthSpecified">
                <xsl:call-template name="TaggedAttributeExists">
                  <xsl:with-param name="tagName" select="'length'"/>
                </xsl:call-template>
              </xsl:variable>
              <xsl:if test="$dataType='string' and $LengthSpecified='false'">
                <xsl:attribute name="length">255</xsl:attribute>
              </xsl:if>
              <xsl:call-template name="WriteReadOnlyAttribute">
                <xsl:with-param name="attrName" select="'readOnly'"/>
              </xsl:call-template>

              <xsl:call-template name="WriteTaggedDocumentation"/>
            </xsl:element>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:when>

      <xsl:when test="//Foundation.Core.Class[@xmi.id = $attrTypeId]">

        <xsl:message>XmiToFdo: mapping XMI Attribute '<xsl:value-of select="Foundation.Core.ModelElement.name"/>' to FDO Object Property '<xsl:value-of select="$fdoPropertyName"/>'</xsl:message>

        <xsl:element name="ObjectProperty">
          <xsl:variable name="classTypeId" select="$attrTypeId"/>
          <xsl:variable name="classItself" select="//Foundation.Core.Class[@xmi.id = $classTypeId]"/>
          <xsl:variable name="maxMultiplicity">
            <xsl:for-each select="Foundation.Core.StructuralFeature.multiplicity">
              <xsl:call-template name="GetMaxMultiplicity"/>
            </xsl:for-each>
          </xsl:variable>
          <xsl:variable name="isOrdered" select="Foundation.Core.ModelElement.taggedValue/Foundation.Extension_Mechanisms.TaggedValue[Foundation.Extension_Mechanisms.TaggedValue.tag = 'orderType']/Foundation.Extension_Mechanisms.TaggedValue.value[text()='ascending' or text()='descending']"/>

          <xsl:attribute name="name">
            <xsl:value-of select="$fdoPropertyName"/>
          </xsl:attribute>
          <xsl:attribute name="class">
            <xsl:call-template name="FormatName">
              <xsl:with-param name="inName" select="$classItself/Foundation.Core.ModelElement.name"/>
            </xsl:call-template>
          </xsl:attribute>
          <xsl:attribute name="classSchema">
            <xsl:call-template name="FormatName">
              <xsl:with-param name="inName">
                <xsl:for-each select="$classItself">
                  <xsl:call-template name="GetSchemaNameForClassNode"/>
                </xsl:for-each>
              </xsl:with-param>
            </xsl:call-template>
          </xsl:attribute>
          <xsl:choose>
            <xsl:when test="$maxMultiplicity = 'm' and $isOrdered">
              <xsl:attribute name="objectType">orderedcollection</xsl:attribute>
            </xsl:when>
            <xsl:when test="$maxMultiplicity = 'm'">
              <xsl:attribute name="objectType">collection</xsl:attribute>
            </xsl:when>
            <xsl:otherwise>
              <xsl:attribute name="objectType">value</xsl:attribute>
            </xsl:otherwise>
          </xsl:choose>
          <xsl:call-template name="WriteTaggedAttributes"/>

          <xsl:call-template name="WriteTaggedDocumentation"/>
        </xsl:element>
      </xsl:when>

    </xsl:choose>
  </xsl:template>


  <!-- The Context Node when this template is called should be a Foundation.Core.AssociationEnd node -->
  <xsl:template name="WriteAssociationProperty">
    <!-- Note: it's a little awkward determine the current node's position relative to parent since position() always returns 1: -->
    <xsl:variable name="posA">
      <xsl:variable name="selfNode" select="."/>
      <xsl:choose>
        <xsl:when test="../child::*[position()=1 and self::node()=$selfNode]">1</xsl:when>
        <xsl:otherwise>2</xsl:otherwise>
      </xsl:choose>
    </xsl:variable>
    <xsl:variable name="posB" select="3 - $posA"/>

    <xsl:variable name="primId" select="../Foundation.Core.AssociationEnd[position() = $posB]/Foundation.Core.AssociationEnd.type/child::*[self::Foundation.Core.Class or self::Foundation.Core.Classifier]/@xmi.idref"/>
    <xsl:variable name="assocName" select="../Foundation.Core.AssociationEnd[position() = $posB]/Foundation.Core.ModelElement.name"/>
    <xsl:variable name="reverseAssocName" select="../Foundation.Core.AssociationEnd[position() = $posA]/Foundation.Core.ModelElement.name"/>
    <xsl:variable name="multiplicity">
      <xsl:for-each select="Foundation.Core.AssociationEnd.multiplicity">
        <xsl:call-template name="GetMaxMultiplicity"/>
      </xsl:for-each>
    </xsl:variable>
    <xsl:variable name="reverseMultiplicity">
      <xsl:for-each select="../Foundation.Core.AssociationEnd[position() = $posB]/Foundation.Core.AssociationEnd.multiplicity">
        <xsl:call-template name="GetMinMultiplicity"/>
      </xsl:for-each>
    </xsl:variable>
    <xsl:variable name="aggregation" select="Foundation.Core.AssociationEnd.aggregation/@xmi.value"/>
    <xsl:variable name="reverseAggregation" select="../Foundation.Core.AssociationEnd[position() = $posB]/Foundation.Core.AssociationEnd.aggregation/@xmi.value"/>

    <xsl:if test="$reverseMultiplicity='1' and not ($multiplicity='1' and $aggregation='composite') and ($posA='1' or $multiplicity='m')">

      <xsl:for-each select="//Foundation.Core.Namespace.ownedElement/child::*[(@xmi.id = $primId) and (self::Foundation.Core.Class or self::Foundation.Core.Classifier)]">

        <xsl:variable name="fdoAssocPropertyName">
          <xsl:call-template name="FormatName">
            <xsl:with-param name="inName" select="$assocName"/>
          </xsl:call-template>
        </xsl:variable>

        <xsl:message>XmiToFdo: mapping XMI Association End '<xsl:value-of select="$assocName"/>' to FDO Association Property '<xsl:value-of select="$fdoAssocPropertyName"/>'</xsl:message>

        <xsl:element name="AssociationProperty">
          <xsl:attribute name="name">
            <xsl:value-of select="$fdoAssocPropertyName"/>
          </xsl:attribute>
          <xsl:attribute name="reverseName">
            <xsl:call-template name="FormatName">
              <xsl:with-param name="inName" select="$reverseAssocName"/>
            </xsl:call-template>
          </xsl:attribute>
          <xsl:attribute name="associatedClass">
            <xsl:call-template name="FormatName">
              <xsl:with-param name="inName" select="Foundation.Core.ModelElement.name"/>
            </xsl:call-template>
          </xsl:attribute>
          <xsl:attribute name="associatedClassSchema">
            <xsl:call-template name="FormatName">
              <xsl:with-param name="inName">
                <xsl:call-template name="GetSchemaNameForClassId">
                  <xsl:with-param name="classid" select="@xmi.id"/>
                </xsl:call-template>
              </xsl:with-param>
            </xsl:call-template>
          </xsl:attribute>
          <xsl:attribute name="multiplicity">
            <xsl:value-of select="$multiplicity"/>
          </xsl:attribute>
          <!--<xsl:attribute name="reverseMultiplicity"><xsl:value-of select="$reverseMultiplicity"/></xsl:attribute>-->
          <xsl:call-template name="WriteReadOnlyAttribute">
            <xsl:with-param name="attrName" select="'isReadonly'"/>
          </xsl:call-template>


          <xsl:element name="IdentityProperties">
            <xsl:attribute name="default">true</xsl:attribute>
          </xsl:element>
          <xsl:element name="IdentityReverseProperties">
            <xsl:attribute name="default">true</xsl:attribute>
          </xsl:element>
        </xsl:element>
      </xsl:for-each>
    </xsl:if>
  </xsl:template>


  <!-- When calling this template, the context node should be a Class node. -->
  <xsl:template name="GetBaseClassId">
    <xsl:if test="Foundation.Core.GeneralizableElement.generalization/Foundation.Core.Generalization">
      <xsl:variable name="genId" select="Foundation.Core.GeneralizableElement.generalization/Foundation.Core.Generalization/@xmi.idref"/>
      <xsl:for-each select="//Foundation.Core.Namespace.ownedElement/Foundation.Core.Generalization[@xmi.id = $genId]">
          <xsl:choose>
            <!-- Visio uses a Foundation.Core.Generalization.supertype/Foundation.Core.Class node to point to parent class: -->
            <xsl:when test="Foundation.Core.Generalization.supertype">
              <xsl:value-of select="Foundation.Core.Generalization.supertype/Foundation.Core.Class/@xmi.idref"/>
            </xsl:when>
            <!-- ArgoUml uses a Foundation.Core.Generalization.parent/Foundation.Core.GeneralizableElement node to point to parent class: -->
            <xsl:when test="Foundation.Core.Generalization.parent">
              <xsl:value-of select="Foundation.Core.Generalization.parent/Foundation.Core.GeneralizableElement/@xmi.idref"/>
            </xsl:when>
          </xsl:choose>
      </xsl:for-each>
    </xsl:if>
  </xsl:template>


  <xsl:template name="WriteBaseClassAttributes">
    <xsl:variable name="baseClassId"><xsl:call-template name="GetBaseClassId"/></xsl:variable>
    <xsl:for-each select="//Foundation.Core.Namespace.ownedElement/Foundation.Core.Class[@xmi.id = $baseClassId]">
      <xsl:attribute name="baseClass">
        <xsl:call-template name="FormatName">
          <xsl:with-param name="inName" select="Foundation.Core.ModelElement.name"/>
        </xsl:call-template>
      </xsl:attribute>
      <xsl:attribute name="baseSchema">
        <xsl:call-template name="FormatName">
          <xsl:with-param name="inName">
            <xsl:call-template name="GetSchemaNameForClassNode"/>
          </xsl:with-param>
        </xsl:call-template>
      </xsl:attribute>
    </xsl:for-each>
  </xsl:template>

  
  <xsl:template name="GetClassType">
    <xsl:variable name="testclassid" select="@xmi.id"/>
    <xsl:variable name="classStereotype">
      <xsl:choose>
        <!-- Visio stores stereotype instances within the ModelElement itself: -->
        <xsl:when test="Foundation.Core.ModelElement.stereotype/Foundation.Extension_Mechanisms.Stereotype/@xmi.idref">
          <xsl:value-of select="Foundation.Core.ModelElement.stereotype/Foundation.Extension_Mechanisms.Stereotype/@xmi.idref"/>
        </xsl:when>
        <!-- ArgoUML stores stereotype instances within the Stereotype itself: -->
        <xsl:when test="//Foundation.Core.Namespace.ownedElement/Foundation.Extension_Mechanisms.Stereotype/Foundation.Extension_Mechanisms.Stereotype.extendedElement/Foundation.Core.ModelElement/@xmi.idref = $testclassid">
          <xsl:value-of select="//Foundation.Core.Namespace.ownedElement/Foundation.Extension_Mechanisms.Stereotype/@xmi.id"/>
        </xsl:when>
      </xsl:choose>
    </xsl:variable>
    <xsl:choose>
      <xsl:when test="$classStereotype=$featStereo">FeatureClass</xsl:when>
      <!-- Network class types; not yet supported: 
      <xsl:when test="$classStereotype=$networkStereo">NetworkClass</xsl:when>
      <xsl:when test="$classStereotype=$networkLayerStereo">NetworkLayerClass</xsl:when>
      <xsl:when test="$classStereotype=$networkLinkFeatureStereo">NetworkLinkFeatureClass</xsl:when>
      <xsl:when test="$classStereotype=$networkNodeFeatureStereo">NetworkNodeFeatureClass</xsl:when>
      -->
      <xsl:otherwise>
        <xsl:variable name="baseClassId"><xsl:call-template name="GetBaseClassId"/></xsl:variable>
        <xsl:choose>
          <xsl:when test="//Foundation.Core.Namespace.ownedElement/Foundation.Core.Class[@xmi.id = $baseClassId]">
            <xsl:for-each select="//Foundation.Core.Namespace.ownedElement/Foundation.Core.Class[@xmi.id = $baseClassId]">
              <xsl:call-template name="GetClassType"/>
            </xsl:for-each>
          </xsl:when>
          <xsl:otherwise>Class</xsl:otherwise>
        </xsl:choose>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>


  <xsl:template name="GetGeometryProperty">
    <xsl:choose>
      <xsl:when test="Foundation.Core.ModelElement.taggedValue/Foundation.Extension_Mechanisms.TaggedValue[Foundation.Extension_Mechanisms.TaggedValue.tag='geometryProperty']">
        <xsl:value-of select="Foundation.Core.ModelElement.taggedValue/Foundation.Extension_Mechanisms.TaggedValue[Foundation.Extension_Mechanisms.TaggedValue.tag='geometryProperty']/Foundation.Extension_Mechanisms.TaggedValue.value"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:variable name="baseClassId"><xsl:call-template name="GetBaseClassId"/></xsl:variable>
        <xsl:if test="$baseClassId">
          <xsl:for-each select="//Foundation.Core.Namespace.ownedElement/Foundation.Core.Class[@xmi.id = $baseClassId]">
            <xsl:call-template name="GetGeometryProperty"/>
          </xsl:for-each>
        </xsl:if>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>


  <xsl:template name="getDataType">
    <xsl:param name="xmiType"/>
    <xsl:variable name="lcaseXmiType" select="translate($xmiType,'ABCDEFGHIJKLMNOPQRSTUVWXYZ','abcdefghijklmnopqrstuvwxyz')"/>
    <xsl:choose>
      <!-- Data types: -->
      <xsl:when test="$lcaseXmiType='bool'"><xsl:value-of select="'boolean'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='boolean'"><xsl:value-of select="'boolean'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='byte'"><xsl:value-of select="'byte'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='datetime'"><xsl:value-of select="'datetime'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='date'"><xsl:value-of select="'datetime'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='time'"><xsl:value-of select="'datetime'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='decimal'"><xsl:value-of select="'decimal'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='bigdecimal'"><xsl:value-of select="'decimal'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='currency'"><xsl:value-of select="'decimal'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='double'"><xsl:value-of select="'double'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='long double'"><xsl:value-of select="'double'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='sbyte'"><xsl:value-of select="'int16'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='short'"><xsl:value-of select="'int16'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='signed short'"><xsl:value-of select="'int16'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='int16'"><xsl:value-of select="'int16'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='unsigned short'"><xsl:value-of select="'int32'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='ushort'"><xsl:value-of select="'int32'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='int'"><xsl:value-of select="'int32'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='signed int'"><xsl:value-of select="'int32'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='integer'"><xsl:value-of select="'int32'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='int32'"><xsl:value-of select="'int32'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='unsigned int'"><xsl:value-of select="'int64'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='uint'"><xsl:value-of select="'int64'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='long'"><xsl:value-of select="'int64'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='signed long'"><xsl:value-of select="'int64'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='unsigned long'"><xsl:value-of select="'int64'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='ulong'"><xsl:value-of select="'int64'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='biginteger'"><xsl:value-of select="'int64'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='int64'"><xsl:value-of select="'int64'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='float'"><xsl:value-of select="'single'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='single'"><xsl:value-of select="'single'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='char'"><xsl:value-of select="'string'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='signed char'"><xsl:value-of select="'string'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='unsigned char'"><xsl:value-of select="'string'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='wchar_t'"><xsl:value-of select="'string'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='string'"><xsl:value-of select="'string'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='url'"><xsl:value-of select="'string'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='bstr'"><xsl:value-of select="'string'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='uninterpreted'"><xsl:value-of select="'blob'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='blob'"><xsl:value-of select="'blob'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='clob'"><xsl:value-of select="'clob'"/></xsl:when>

      <!-- Geometry types: -->
      <xsl:when test="$lcaseXmiType='geometric'"><xsl:value-of select="'Geometric'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='geometry'"><xsl:value-of select="'Geometric'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='vector'"><xsl:value-of select="'Geometric'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='point'"><xsl:value-of select="'Geometric'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='line'"><xsl:value-of select="'Geometric'"/></xsl:when>
      <xsl:when test="$lcaseXmiType='polygon'"><xsl:value-of select="'Geometric'"/></xsl:when>

      <!-- Raster types: -->
      <xsl:when test="$lcaseXmiType='raster'"><xsl:value-of select="'Raster'"/></xsl:when>

      <xsl:otherwise><xsl:value-of select="$xmiType"/></xsl:otherwise>
    </xsl:choose>
  </xsl:template>


  <xsl:template name="FormatName">
    <xsl:param name="inName"/>
    <xsl:choose>
      <!-- Don't alter the original class names, since this is provider-specific:
      <xsl:when test="$inName='Feature'">
        <xsl:value-of select="'Feature_'"/>
      </xsl:when>
      -->
      <xsl:otherwise>
        <xsl:value-of select="translate($inName,' ','_')"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>


  <xsl:template name="FormatMultiplicity">
    <xsl:param name="inMultiplicity"/>
    <xsl:choose>
      <!-- ArgoUml uses '-1' to mean 'many' -->
      <xsl:when test="$inMultiplicity='-1'">m</xsl:when>
      <!-- Visio uses '*' to mean 'many' -->
      <xsl:when test="$inMultiplicity='*'">m</xsl:when>
      <!-- Pass through other multiplicity values (0,1,2,etc): -->
      <xsl:otherwise><xsl:value-of select="$inMultiplicity"/></xsl:otherwise>
    </xsl:choose>
  </xsl:template>


  <xsl:template name="GetSchemaNameForClassId">
    <xsl:param name="classid"/>
    <xsl:for-each select="//Foundation.Core.Namespace.ownedElement/Foundation.Core.Class[@xmi.id = $classid]">
      <xsl:call-template name="GetSchemaNameForClassNode"/>
    </xsl:for-each>
  </xsl:template>


  <xsl:template name="GetSchemaNameForClassNode">
    <xsl:value-of select="(ancestor::Model_Management.Package|ancestor::Model_Management.Subsystem|ancestor::Model_Management.Model)[not((ancestor::Model_Management.Package|ancestor::Model_Management.Subsystem|ancestor::Model_Management.Model))]/Foundation.Core.ModelElement.name"/>
  </xsl:template>


  <xsl:template name="GetStereotypedAssociationEnd">
    <xsl:param name="stereotypeid"/>
    <xsl:variable name="classid" select="@xmi.id"/>
    <!-- TODO: handle ArgoUML's way of stereotyping associations; not critical for now since this is only used for Networking, which is not supported yet -->
    <xsl:value-of select="//Foundation.Core.Namespace.ownedElement/Foundation.Core.Association/Foundation.Core.Association.connection/Foundation.Core.AssociationEnd[Foundation.Core.AssociationEnd.type/child::*[self::Foundation.Core.Class or self::Foundation.Core.Classifier]/@xmi.idref=$classid and ../../@xmi.id=/descendant-or-self::Foundation.Core.Namespace.ownedElement/Foundation.Extension_Mechanisms.Stereotype[@xmi.id = $stereotypeid]/Foundation.Extension_Mechanisms.Stereotype.extendedElement/Foundation.Core.ModelElement/@xmi.idref]"/>
  </xsl:template>


  <xsl:template name="TaggedAttributeExists">
    <xsl:param name="tagName"/>
    <xsl:variable name="exists" select="Foundation.Core.ModelElement.taggedValue/Foundation.Extension_Mechanisms.TaggedValue[Foundation.Extension_Mechanisms.TaggedValue.tag = $tagName]"/>
    <xsl:copy-of select="count($exists) > 0"/>
  </xsl:template>


</xsl:stylesheet>
